/* poplaun v0.1 (February 2017)
 * Copyright (C) 2017 Norbert de Jonge <mail@norbertdejonge.nl>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see [ www.gnu.org/licenses/ ].
 *
 * To properly read this code, set your program's tab stop to: 2.
 */

/*========== Includes ==========*/
#include <gtk/gtk.h>

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/HTMLparser.h>

#include <zip.h>

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#include <windows.h>
#define GetMenu POPLAUN_GetMenu
#endif
/*========== Includes ==========*/

/*========== Defines ==========*/
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#define SLASH "\\"
#define HERE ""
#else
#define SLASH "/"
#define HERE "./"
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define PROG_NAME "poplaun"
#define PROG_VERSION "0.1"
#define PROG_AUTHOR_MAIN "Norbert de Jonge"
#define PROG_MONTH_YEAR "February 2017"
#define PROG_COPY_YEAR_LAST "2017"
#define PROG_COPY_YEARS "2017"
#define PROG_COPY_LAST "Copyright (C) " PROG_COPY_YEAR_LAST \
	" " PROG_AUTHOR_MAIN
#define PROG_COPY_FULL "Copyright (C) " PROG_COPY_YEARS \
	" " PROG_AUTHOR_MAIN
#define PROG_DESC "A tool to download and launch Prince of Persia mods" \
	" by parsing popot.org XML files."
#define PROG_WEBSITE "http://www.norbertdejonge.nl/poplaun/"

#define EXIT_NORMAL 0
#define EXIT_ERROR 1
#define MODS_DIR "mods"
#define XML_DIR "xml"
#define XML_MODSU "modsu.xml"
#define XML_MODS "mods.xml"
#define MAX_LOCATION 500
#define MAX_NODENAME 100
#define DATE 20
#define MAX_MODS 1000
#define MAX_COMMENTS 100
#define MAX_REPLAYS 400
#define TEXT_LARGE 5000
#define TEXT_HUGE 50000
#define MAX_WRITE 200
#define NO_SPECIFIC_LEVEL "full"
#define BIN_DOSBOX "/usr/bin/dosbox"
#define BIN_ZSNES "/usr/bin/zsnes"
/*========== Defines ==========*/

GtkWidget *gGeneralLabel;
GtkWidget *gGeneralLink;
GtkWidget *gShotsGrid;
GtkWidget *gShotsImage[10 + 2];
GtkWidget *gCommentsLabel;
GtkWidget *gReplaysLabel;
GtkWidget *gMainLabel2;
GtkWidget *gMainWindow;
GtkWidget *gMainCButton;
GtkWidget *gMainButtonGet;
GtkWidget *gMainButtonPlay;
GtkListStore *gMainListStore;
GtkTreeModel *gMainTreeModel;
GtkWidget *gMainEntry;
GtkWidget *gLevelCombo;
GtkTreeModel *gLevelModel;
GdkPixbuf *gIconDefault;
GtkWidget *gMainTabLabel2;
GtkWidget *gMainTabLabel3;
GtkWidget *gMainTabLabel4;
GtkWidget *gWaitWindow;
/***/
int iModActive;
int iModShowing;
int iModLevel;
int iModCheat;
int iDebug;
char sNodeName[MAX_NODENAME + 2];
int iDateWhen;
char sModsUDateOld[DATE + 2];
char sModsUDateNew[DATE + 2];
int iNrModsOld, iNrModsNew;
int iSpecialBlock;
int iNrComments;
int iNrReplays;
char sPNG[MAX_LOCATION + 2];
char sTXT[MAX_LOCATION + 2];
char sModDir[MAX_LOCATION + 2];
struct stat stStatus;
int iWait;

/*** Old ***/
char arOldMods_S_Number[MAX_MODS + 2][7 + 2];
char arOldMods_S_Last_Update[MAX_MODS + 2][DATE + 2];
char arOldMods_S_Last_Update_ZIP[MAX_MODS + 2][DATE + 2];

/*** New ***/
char arMods_S_Name[MAX_MODS + 2][100 + 2];
int arMods_I_Number[MAX_MODS + 2];
char arMods_S_Number[MAX_MODS + 2][7 + 2];
int arMods_I_Creation_Year[MAX_MODS + 2];
int arMods_I_PoP_Version[MAX_MODS + 2];
char arMods_S_PoP_Version[MAX_MODS + 2][20 + 2];
char arMods_S_Description[MAX_MODS + 2][200 + 2];
int arMods_I_Starting_Minutes[MAX_MODS + 2];
int arMods_I_Author1_Id[MAX_MODS + 2];
char arMods_S_Author1_Name[MAX_MODS + 2][50 + 2];
int arMods_I_Author1_Type[MAX_MODS + 2];
int arMods_I_Author2_Id[MAX_MODS + 2];
char arMods_S_Author2_Name[MAX_MODS + 2][50 + 2];
int arMods_I_Author2_Type[MAX_MODS + 2];
int arMods_I_Author3_Id[MAX_MODS + 2];
char arMods_S_Author3_Name[MAX_MODS + 2][50 + 2];
int arMods_I_Author3_Type[MAX_MODS + 2];
int arMods_I_Altered_Graphics[MAX_MODS + 2];
int arMods_I_Altered_Audio[MAX_MODS + 2];
char arMods_S_Executable_DOSSNES[MAX_MODS + 2][100 + 2];
char arMods_S_Executable_SDLPoP[MAX_MODS + 2][100 + 2];
char arMods_S_Executable_MININIM[MAX_MODS + 2][100 + 2];
char arMods_S_Cheat_Code[MAX_MODS + 2][100 + 2];
double arMods_F_Enjoyment[MAX_MODS + 2];
double arMods_F_Difficulty[MAX_MODS + 2];
int arMods_I_Levels_Changed[MAX_MODS + 2];
int arMods_I_Enjoyment_Votes[MAX_MODS + 2];
int arMods_I_Difficulty_Votes[MAX_MODS + 2];
int arMods_I_Screenshots[MAX_MODS + 2];
int arMods_I_Comments[MAX_MODS + 2];
int arMods_I_Replays[MAX_MODS + 2];
char arMods_S_Last_Update[MAX_MODS + 2][DATE + 2];
char arMods_S_Last_Update_ZIP[MAX_MODS + 2][DATE + 2];
/***/
char arComments_S_Nick[MAX_MODS + 2][MAX_COMMENTS + 2][40 + 2];
char arComments_S_Added[MAX_MODS + 2][MAX_COMMENTS + 2][DATE + 2];
char arComments_S_Comment[MAX_MODS + 2][MAX_COMMENTS + 2][TEXT_LARGE + 2];
/***/
int arReplays_I_Level[MAX_MODS + 2][MAX_REPLAYS + 2];
char arReplays_S_Level[MAX_MODS + 2][MAX_REPLAYS + 2][20 + 2];
char arReplays_S_Program[MAX_MODS + 2][MAX_REPLAYS + 2][10 + 2];
char arReplays_S_Nick[MAX_MODS + 2][MAX_REPLAYS + 2][40 + 2];
char arReplays_S_Added[MAX_MODS + 2][MAX_REPLAYS + 2][DATE + 2];
char arReplays_S_Comment[MAX_MODS + 2][MAX_REPLAYS + 2][100 + 2];
char arReplays_S_URL[MAX_MODS + 2][MAX_REPLAYS + 2][60 + 2];

void CreateDir (char *sDir);
void GetXML (char *sFile);
void ModsU (char *sLocation);
void ModsUNode (xmlNode *node);
int DateTimeToPosix (char *sDateTime);
void Mods (char *sLocation, int iFull);
void ModsNode (xmlNode *node, int iFull);
void StoreS (char *sName, char *sValue, char *sTarget, int iSize);
void StoreI (char *sName, int iValue, int *iTarget);
void StoreF (char *sName, double dValue, double *dTarget);
void MainWindow (void);
void Quit (GtkWidget *widget, gpointer user_data);
gboolean QuitEvent (GtkWidget *widget, GdkEvent *event, gpointer user_data);
void UpdateList (void);
void UpdateShots (void);
void Get (GtkButton *button, gpointer user_data);
void Play (GtkButton *button, gpointer user_data);
void Cheat (GtkButton *button, gpointer user_data);
void SelectRow (GtkTreeSelection *gList, gpointer user_data);
void ShowMod (void);
void GetMenu (GtkWidget **widget, int iMenu, int iX);
void AddMenuItem (GtkWidget **gMenu, char *sEntry,
	char *sIcon, char *sFunc, int iX);
void About (GtkMenuItem *menuitem, gpointer user_data);
void ShowUsage (void);
gboolean GlobalSearch (GtkTreeModel *model, gint column,
	const gchar *key, GtkTreeIter *iter, gpointer search_data);
void Refresh (GtkButton *button, gpointer user_data);
void ParseXML (void);
void UpdateButtons (void);
void GetMod (char *sNumber);
void GetShots (char *sNumber);
void UnzipMod (char *sNumber);
void InsertLevels (void);
void Level (GtkComboBox *widget, gpointer user_data);
void PleaseWait (void);
void Done (GObject *source_object, GAsyncResult *res, gpointer user_data);
void CreateBatchFile (char *sNumber);
void RunGame (char *sNumber);
void WarnIfNotExists (char *sFile);

/*****************************************************************************/
int main (int argc, char *argv[])
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iArgLoop;

	/*** Defaults. ***/
	iModActive = 0;
	iModShowing = 0;
	iModLevel = 0;
	iModCheat = 0;
	iDebug = 0;
	iWait = 0;

	if (argc > 1)
	{
		for (iArgLoop = 1; iArgLoop <= argc - 1; iArgLoop++)
		{
			if ((strcmp (argv[iArgLoop], "-h") == 0) ||
				(strcmp (argv[iArgLoop], "-?") == 0) ||
				(strcmp (argv[iArgLoop], "--help") == 0))
			{
				ShowUsage();
			}
			else if ((strcmp (argv[iArgLoop], "-v") == 0) ||
				(strcmp (argv[iArgLoop], "--version") == 0))
			{
				printf ("%s v%s (%s)\n", PROG_NAME, PROG_VERSION, PROG_MONTH_YEAR);
				exit (EXIT_NORMAL);
			}
			else if ((strcmp (argv[iArgLoop], "-d") == 0) ||
				(strcmp (argv[iArgLoop], "--debug") == 0))
			{
				iDebug = 1;
			}
			else
			{
				ShowUsage();
			}
		}
	}

	/*** Initialize libraries. ***/
	gtk_init (&argc, &argv);
	LIBXML_TEST_VERSION

	/*** In debug mode, show the GTK+ version. ***/
	if (iDebug == 1)
	{
		printf ("[ INFO ] Compiled with GTK+ %u.%u.%u,"
			" linked with GTK+ %u.%u.%u.\n",
			GTK_MAJOR_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION,
			gtk_major_version, gtk_minor_version, gtk_micro_version);
	}

	/*** Preload default window icon. ***/
	snprintf (sPNG, MAX_LOCATION, "png%s%s%s", SLASH, PROG_NAME, "_icon.png");
	gIconDefault = gdk_pixbuf_new_from_file (sPNG, NULL);
	gtk_window_set_default_icon (gIconDefault);

	/*** Create directories. ***/
	CreateDir (MODS_DIR);
	CreateDir (XML_DIR);

	MainWindow();

	ParseXML();
	UpdateList();

	gtk_main();
	Quit (NULL, NULL);
	return EXIT_NORMAL;
}
/*****************************************************************************/
void CreateDir (char *sDir)
/*****************************************************************************/
{
	if (stat (sDir, &stStatus) == -1)
	{
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
mkdir (sDir);
#else
mkdir (sDir, 0700);
#endif
	}
}
/*****************************************************************************/
void GetXML (char *sFile)
/*****************************************************************************/
{
	char sFrom[MAX_LOCATION + 2], sTo[MAX_LOCATION + 2];
	GFile *remote, *local;
	gboolean bCopied;

	snprintf (sFrom, MAX_LOCATION, "http://www.popot.org/xml/%s", sFile);
	remote = g_file_new_for_uri (sFrom);
	snprintf (sTo, MAX_LOCATION, "%s%s%s", XML_DIR, SLASH, sFile);
	local = g_file_new_for_path (sTo);

	if (iDebug == 1)
	{
		printf ("[ INFO ] Copying \"%s\" to \"%s\".\n", sFrom, sTo);
	}

	bCopied = g_file_copy (remote, local, G_FILE_COPY_OVERWRITE,
		NULL, NULL, NULL, NULL);
	if (bCopied == FALSE)
	{
		printf ("[FAILED] Could not retrieve \"%s\": %s!\n",
			sFile, strerror (errno));
		exit (EXIT_ERROR);
	}
}
/*****************************************************************************/
void ModsU (char *sLocation)
/*****************************************************************************/
{
	xmlDoc *doc = NULL;
	xmlNode *root_element = NULL;

	doc = xmlReadFile (sLocation, NULL, HTML_PARSE_NOBLANKS|HTML_PARSE_NOERROR|
		HTML_PARSE_NOWARNING|HTML_PARSE_NONET);

	if (doc == NULL)
	{
		printf ("[FAILED] Could not read XML file!\n");
		exit (EXIT_ERROR);
	}

	root_element = xmlDocGetRootElement (doc);
	if (root_element == NULL)
	{
		printf ("[FAILED] Could not get root element!\n");
		exit (EXIT_ERROR);
	}

	ModsUNode (root_element);

	xmlFreeDoc (doc);
}
/*****************************************************************************/
void ModsUNode (xmlNode *node)
/*****************************************************************************/
{
	xmlNode *cur_node = NULL;

	if (node == NULL) { return; }

	for (cur_node = node; cur_node != NULL; cur_node = cur_node->next)
	{
		switch (cur_node->type)
		{
			case XML_ELEMENT_NODE:
				snprintf (sNodeName, MAX_NODENAME, "%s", cur_node->name);
				break;
			case XML_TEXT_NODE:
				if (strcmp (sNodeName, "dt_last_update") == 0)
				{
					switch (iDateWhen)
					{
						case 0: /*** old ***/
							snprintf (sModsUDateOld, DATE, "%s", cur_node->content);
							break;
						case 1: /*** new ***/
							snprintf (sModsUDateNew, DATE, "%s", cur_node->content);
							break;
					}
				}
				break;
			default: break;
		}
		ModsUNode (cur_node->children);
	}
}
/*****************************************************************************/
int DateTimeToPosix (char *sDateTime)
/*****************************************************************************/
{
	unsigned int iYear, iMonth, iDay, iHour, iMinute, iSecond;
	struct tm tmTime;
	int iPosix;

	sscanf (sDateTime, "%4u-%2u-%2u %2u:%2u:%2u", &iYear, &iMonth, &iDay, &iHour,
		&iMinute, &iSecond);
	tmTime.tm_isdst = -1;
	tmTime.tm_year = iYear - 1900;
	tmTime.tm_mon = iMonth - 1;
	tmTime.tm_mday = iDay;
	tmTime.tm_hour = iHour;
	tmTime.tm_min = iMinute;
	tmTime.tm_sec = iSecond;

	iPosix = mktime (&tmTime);

	return (iPosix);
}
/*****************************************************************************/
void Mods (char *sLocation, int iFull)
/*****************************************************************************/
{
	xmlDoc *doc = NULL;
	xmlNode *root_element = NULL;

	doc = xmlReadFile (sLocation, NULL, HTML_PARSE_NOBLANKS|HTML_PARSE_NOERROR|
		HTML_PARSE_NOWARNING|HTML_PARSE_NONET);

	if (doc == NULL)
	{
		printf ("[FAILED] Could not read XML file!\n");
		exit (EXIT_ERROR);
	}

	root_element = xmlDocGetRootElement (doc);
	if (root_element == NULL)
	{
		printf ("[FAILED] Could not get root element!\n");
		exit (EXIT_ERROR);
	}

	iSpecialBlock = 0;
	ModsNode (root_element, iFull);

	xmlFreeDoc (doc);
}
/*****************************************************************************/
void ModsNode (xmlNode *node, int iFull)
/*****************************************************************************/
{
	xmlNode *cur_node = NULL;
	xmlChar *attr = NULL;
	int iValue;
	double dValue;
	char sValue[TEXT_LARGE + 2];

	if (node == NULL) { return; }

	for (cur_node = node; cur_node != NULL; cur_node = cur_node->next)
	{
		switch (cur_node->type)
		{
			case XML_ELEMENT_NODE:
				snprintf (sNodeName, MAX_NODENAME, "%s", cur_node->name);
				if (strcmp (sNodeName, "mod") == 0)
				{
					attr = xmlGetProp (cur_node, (xmlChar *)"id");
					switch (iFull)
					{
						case 0:
							iNrModsOld++;
							break;
						case 1:
							if (iDebug == 1)
							{
								printf ("[ INFO ] Parsing mod \"%s\".\n", attr);
							}
							iNrModsNew++;
							break;
					}
				}
				if (strcmp (sNodeName, "comments") == 0)
				{
					iSpecialBlock = 1;
					iNrComments = 0;
				}
				if (strcmp (sNodeName, "comment") == 0) { iNrComments++; }
				if (strcmp (sNodeName, "i_replays") == 0) { iSpecialBlock = 0; }
				if (strcmp (sNodeName, "replays") == 0)
				{
					iSpecialBlock = 2;
					iNrReplays = 0;
				}
				if (strcmp (sNodeName, "replay") == 0) { iNrReplays++; }
				if (strcmp (sNodeName, "dt_last_update") == 0) { iSpecialBlock = 0; }
				break;
			case XML_TEXT_NODE:
				iValue = atoi ((char *)cur_node->content);
				dValue = atof ((char *)cur_node->content);
				snprintf (sValue, TEXT_LARGE, "%s", cur_node->content);
				if (iSpecialBlock == 0)
				{
					if (iFull == 1)
					{
						StoreS ("s_name", sValue,
							arMods_S_Name[iNrModsNew], 100);
						StoreI ("i_number", iValue,
							&arMods_I_Number[iNrModsNew]);
						StoreS ("s_number", sValue,
							arMods_S_Number[iNrModsNew], 7);
						StoreI ("i_creation_year", iValue,
							&arMods_I_Creation_Year[iNrModsNew]);
						StoreI ("i_pop_version", iValue,
							&arMods_I_PoP_Version[iNrModsNew]);
						StoreS ("s_pop_version", sValue,
							arMods_S_PoP_Version[iNrModsNew], 20);
						StoreS ("s_description", sValue,
							arMods_S_Description[iNrModsNew], 200);
						StoreI ("i_starting_minutes", iValue,
							&arMods_I_Starting_Minutes[iNrModsNew]);
						StoreI ("i_author1_id", iValue,
							&arMods_I_Author1_Id[iNrModsNew]);
						StoreS ("s_author1_name", sValue,
							arMods_S_Author1_Name[iNrModsNew], 50);
						StoreI ("i_author1_type", iValue,
							&arMods_I_Author1_Type[iNrModsNew]);
						StoreI ("i_author2_id", iValue,
							&arMods_I_Author2_Id[iNrModsNew]);
						StoreS ("s_author2_name", sValue,
							arMods_S_Author2_Name[iNrModsNew], 50);
						StoreI ("i_author2_type", iValue,
							&arMods_I_Author2_Type[iNrModsNew]);
						StoreI ("i_author3_id", iValue,
							&arMods_I_Author3_Id[iNrModsNew]);
						StoreS ("s_author3_name", sValue,
							arMods_S_Author3_Name[iNrModsNew], 50);
						StoreI ("i_author3_type", iValue,
							&arMods_I_Author3_Type[iNrModsNew]);
						StoreI ("i_altered_graphics", iValue,
							&arMods_I_Altered_Graphics[iNrModsNew]);
						StoreI ("i_altered_audio", iValue,
							&arMods_I_Altered_Audio[iNrModsNew]);
						StoreS ("s_executable", sValue,
							arMods_S_Executable_DOSSNES[iNrModsNew], 100);
						StoreS ("s_executable_sdlpop", sValue,
							arMods_S_Executable_SDLPoP[iNrModsNew], 100);
						StoreS ("s_executable_mininim", sValue,
							arMods_S_Executable_MININIM[iNrModsNew], 100);
						StoreS ("s_cheat_code", sValue,
							arMods_S_Cheat_Code[iNrModsNew], 100);
						StoreF ("f_enjoyment", dValue,
							&arMods_F_Enjoyment[iNrModsNew]);
						StoreF ("f_difficulty", dValue,
							&arMods_F_Difficulty[iNrModsNew]);
						StoreI ("i_levels_changed", iValue,
							&arMods_I_Levels_Changed[iNrModsNew]);
						StoreI ("i_enjoyment_votes", iValue,
							&arMods_I_Enjoyment_Votes[iNrModsNew]);
						StoreI ("i_difficulty_votes", iValue,
							&arMods_I_Difficulty_Votes[iNrModsNew]);
						StoreI ("i_screenshots", iValue,
							&arMods_I_Screenshots[iNrModsNew]);
						StoreI ("i_comments", iValue,
							&arMods_I_Comments[iNrModsNew]);
						StoreI ("i_replays", iValue,
							&arMods_I_Replays[iNrModsNew]);
						StoreS ("dt_last_update", sValue,
							arMods_S_Last_Update[iNrModsNew], DATE);
						StoreS ("dt_last_update_zip", sValue,
							arMods_S_Last_Update_ZIP[iNrModsNew], DATE);
					} else {
						StoreS ("s_number", sValue, arOldMods_S_Number[iNrModsOld], 7);
						StoreS ("dt_last_update", sValue,
							arOldMods_S_Last_Update[iNrModsOld], DATE);
						StoreS ("dt_last_update_zip", sValue,
							arOldMods_S_Last_Update_ZIP[iNrModsOld], DATE);
					}
				}
				if (iSpecialBlock == 1) /*** comments ***/
				{
					if (iFull == 1)
					{
						StoreS ("s_nick", sValue,
							arComments_S_Nick[iNrModsNew][iNrComments], 40);
						StoreS ("dt_added", sValue,
							arComments_S_Added[iNrModsNew][iNrComments], DATE);
						StoreS ("s_comment", sValue,
							arComments_S_Comment[iNrModsNew][iNrComments], TEXT_LARGE);
					}
				}
				if (iSpecialBlock == 2) /*** replays ***/
				{
					if (iFull == 1)
					{
						StoreI ("i_level", iValue,
							&arReplays_I_Level[iNrModsNew][iNrReplays]);
						StoreS ("s_level", sValue,
							arReplays_S_Level[iNrModsNew][iNrReplays], 20);
						StoreS ("s_program", sValue,
							arReplays_S_Program[iNrModsNew][iNrReplays], 10);
						StoreS ("s_nick", sValue,
							arReplays_S_Nick[iNrModsNew][iNrReplays], 40);
						StoreS ("dt_added", sValue,
							arReplays_S_Added[iNrModsNew][iNrReplays], DATE);
						StoreS ("s_comment", sValue,
							arReplays_S_Comment[iNrModsNew][iNrReplays], 100);
						StoreS ("s_url", sValue,
							arReplays_S_URL[iNrModsNew][iNrReplays], 60);
					}
				}
				break;
			default: break;
		}
		ModsNode (cur_node->children, iFull);
	}
}
/*****************************************************************************/
void StoreS (char *sName, char *sValue, char *sTarget, int iSize)
/*****************************************************************************/
{
	if (strcmp (sNodeName, sName) == 0)
	{
		snprintf (sTarget, iSize + 2, "%s", sValue);
	}
}
/*****************************************************************************/
void StoreI (char *sName, int iValue, int *iTarget)
/*****************************************************************************/
{
	if (strcmp (sNodeName, sName) == 0)
	{
		*iTarget = iValue;
	}
}
/*****************************************************************************/
void StoreF (char *sName, double dValue, double *dTarget)
/*****************************************************************************/
{
	if (strcmp (sNodeName, sName) == 0)
	{
		*dTarget = dValue;
	}
}
/*****************************************************************************/
void MainWindow (void)
/*****************************************************************************/
{
	GtkWidget *gMainBox;
	GtkWidget *gMainGrid;
	GtkWidget *gMainFrameGrid;
	GtkWidget *gMainTreeView;
	GtkWidget *gMainFrame;
	GtkWidget *gMainTab;
	GtkWidget *gMainTabLabel1;
	GtkWidget *gMainTabGrid1;
	GtkWidget *gMainTabGrid2;
	GtkWidget *gMainTabGrid3;
	GtkWidget *gMainTabGrid4;
	GtkWidget *gMainMenu;
	GtkWidget *gMainLabel1;
	GtkWidget *gMainButtonRefresh;
	GtkWidget *gMainImage;
	GtkWidget *gMainSep;
	/***/
	GtkWidget *gGeneralGrid;
	GtkWidget *gGeneralScroll;
	/***/
	GtkWidget *gShotsScroll;
	/***/
	GtkWidget *gCommentsScroll;
	/***/
	GtkWidget *gReplaysScroll;
	/***/
	GtkWidget *gLevelGrid;
	GtkWidget *gLevelLabel;
	/***/
	GtkTreeViewColumn *gMainTreeViewColumn[2 + 2];
	GtkCellRenderer *renderer, *rendererp;
	GtkWidget *gMainScroll;
	GtkTreeSelection *gSelection;

	/*** Used for looping. ***/
	int iColumnLoop;
	int iShotLoop;

	gchar *ModColumns[3] = {
		"#",
		"", /*** icon ***/
		"Name",
	};

	/*** window ***/
	gMainWindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_size_request (GTK_WIDGET (gMainWindow), 720, 405);
	gtk_window_set_position (GTK_WINDOW (gMainWindow), GTK_WIN_POS_CENTER);
	gtk_window_set_title (GTK_WINDOW (gMainWindow),
		PROG_NAME " v" PROG_VERSION " (" PROG_MONTH_YEAR ")");
	g_signal_connect (gMainWindow, "delete_event",
		G_CALLBACK (QuitEvent), NULL);
	gtk_widget_realize (gMainWindow);

	/*** box ***/
	gMainBox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
	gtk_container_add (GTK_CONTAINER (gMainWindow), gMainBox);
	gtk_widget_show (gMainBox);

	/*** grid ***/
	gMainGrid = gtk_grid_new();
	gtk_widget_show (gMainGrid);
	gtk_box_pack_start (GTK_BOX (gMainBox), gMainGrid, TRUE, TRUE, 0);

	/*** menu ***/
	GetMenu (&gMainMenu, 1, 0);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainMenu, 0, 0, 6, 1);
	g_object_set (gMainMenu, "hexpand", TRUE, NULL);
	gtk_widget_show_all (gMainMenu);

	/*** scroll ***/
	gMainScroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (gMainScroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	g_object_set (gMainScroll, "shadow-type", GTK_SHADOW_IN, NULL);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainScroll, 0, 1, 3, 3);
	g_object_set (gMainScroll, "expand", TRUE, NULL);
	g_object_set (gMainScroll, "margin", 10, NULL);
	gtk_widget_show (gMainScroll);

	/*** tree view (1/2) ***/
	gMainTreeView = gtk_tree_view_new();

	/*** renderer ***/
	renderer = gtk_cell_renderer_text_new();
	rendererp = gtk_cell_renderer_pixbuf_new();
	for (iColumnLoop = 0; iColumnLoop < 3; iColumnLoop++)
	{
		if (iColumnLoop != 1)
		{
			gMainTreeViewColumn[iColumnLoop] =
				gtk_tree_view_column_new_with_attributes (
				ModColumns[iColumnLoop], renderer, "text", iColumnLoop, NULL);
		} else {
			gMainTreeViewColumn[iColumnLoop] =
				gtk_tree_view_column_new_with_attributes (
				ModColumns[iColumnLoop], rendererp, "pixbuf", iColumnLoop, NULL);
		}
		gtk_tree_view_append_column (GTK_TREE_VIEW (gMainTreeView),
			gMainTreeViewColumn[iColumnLoop]);
		gtk_tree_view_column_set_resizable (gMainTreeViewColumn[iColumnLoop],
			FALSE);
		gtk_tree_view_column_set_sizing (gMainTreeViewColumn[iColumnLoop],
			GTK_TREE_VIEW_COLUMN_AUTOSIZE);
		gtk_tree_view_column_set_clickable (gMainTreeViewColumn[iColumnLoop],
			FALSE);
	}

	/*** store ***/
	gMainListStore = gtk_list_store_new (3,
		G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_STRING);

	/*** model ***/
	gMainTreeModel = GTK_TREE_MODEL (gMainListStore);
	gtk_tree_view_set_model (GTK_TREE_VIEW (gMainTreeView), gMainTreeModel);
	g_object_unref (gMainTreeModel);

	/*** tree view (2/2) ***/
	g_object_set (gMainTreeView, "enable-grid-lines",
		GTK_TREE_VIEW_GRID_LINES_HORIZONTAL, NULL);
	gtk_container_add (GTK_CONTAINER (gMainScroll), gMainTreeView);
	gtk_widget_show (gMainTreeView);

	/*** selection ***/
	gSelection = gtk_tree_view_get_selection (GTK_TREE_VIEW (gMainTreeView));
	gtk_tree_selection_set_mode (gSelection, GTK_SELECTION_SINGLE);
	g_signal_connect (gSelection, "changed",
		G_CALLBACK (SelectRow), NULL);

	/*** button ***/
	gMainButtonRefresh = gtk_button_new();
	snprintf (sPNG, MAX_LOCATION, "png%srefresh.png", SLASH);
	gMainImage = gtk_image_new_from_file (sPNG);
	gtk_button_set_image (GTK_BUTTON (gMainButtonRefresh), gMainImage);
	gtk_button_set_always_show_image (GTK_BUTTON (gMainButtonRefresh), TRUE);
	g_signal_connect (gMainButtonRefresh, "clicked",
		G_CALLBACK (Refresh), NULL);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainButtonRefresh, 0, 4, 1, 1);
	gtk_widget_set_margin_start (gMainButtonRefresh, 10);
	gtk_widget_set_margin_bottom (gMainButtonRefresh, 10);
	gtk_widget_show (gMainButtonRefresh);

	/*** label ***/
	gMainLabel1 = gtk_label_new ("Search:");
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainLabel1, 1, 4, 1, 1);
	g_object_set (gMainLabel1, "valign", GTK_ALIGN_CENTER, NULL);
	g_object_set (gMainLabel1, "margin", 10, NULL);
	gtk_widget_set_margin_top (gMainLabel1, 0);
	gtk_widget_show (gMainLabel1);

	/*** entry ***/
	gMainEntry = gtk_entry_new();
	gtk_entry_set_max_length (GTK_ENTRY (gMainEntry), 100);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainEntry, 2, 4, 1, 1);
	g_object_set (gMainEntry, "hexpand", TRUE, NULL);
	gtk_widget_set_margin_bottom (gMainEntry, 10);
	gtk_widget_set_margin_end (gMainEntry, 10);
	/***/
	gtk_tree_view_set_search_entry (GTK_TREE_VIEW (gMainTreeView),
		GTK_ENTRY (gMainEntry));
	gtk_tree_view_set_enable_search (GTK_TREE_VIEW (gMainTreeView), TRUE);
	gtk_tree_view_set_search_column (GTK_TREE_VIEW (gMainTreeView), 2);
	gtk_tree_view_set_search_equal_func (GTK_TREE_VIEW (gMainTreeView),
		GlobalSearch, gMainTreeModel, NULL);
	/***/
	gtk_widget_grab_focus (gMainEntry);
	gtk_widget_show (gMainEntry);

	/*** separator ***/
	gMainSep = gtk_separator_new (GTK_ORIENTATION_VERTICAL);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainSep, 3, 1, 1, 4);
	gtk_widget_show (gMainSep);

	/*** label ***/
	gMainLabel2 = gtk_label_new ("(no mod selected)");
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainLabel2, 4, 1, 2, 1);
	g_object_set (gMainLabel2, "margin", 10, NULL);
	gtk_widget_set_margin_bottom (gMainLabel2, 0);
	gtk_widget_show (gMainLabel2);

	/*** frame ***/
	gMainFrame = gtk_frame_new (NULL);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainFrame, 4, 2, 2, 1);
	g_object_set (gMainFrame, "margin", 10, NULL);
	g_object_set (gMainFrame, "vexpand", TRUE, NULL);
	gtk_widget_show (gMainFrame);

	/*** grid ***/
	gMainFrameGrid = gtk_grid_new();
	gtk_container_add (GTK_CONTAINER (gMainFrame), gMainFrameGrid);
	gtk_widget_show (gMainFrameGrid);

	/*** tab ***/
	gMainTab = gtk_notebook_new();
	gtk_notebook_set_scrollable (GTK_NOTEBOOK (gMainTab), TRUE);
	gtk_notebook_set_tab_pos (GTK_NOTEBOOK (gMainTab), GTK_POS_TOP);
	gtk_grid_attach (GTK_GRID (gMainFrameGrid), gMainTab, 0, 0, 1, 1);
	g_object_set (gMainTab, "margin", 10, NULL);
	gtk_widget_show (gMainTab);

	/*** "General" ***/
	gMainTabLabel1 = gtk_label_new ("General");
	gtk_widget_show (gMainTabLabel1);
	gMainTabGrid1 = gtk_grid_new();
	g_object_set (gMainTabGrid1, "vexpand", TRUE, NULL);
	gtk_widget_show (gMainTabGrid1);
	gtk_notebook_append_page (GTK_NOTEBOOK (gMainTab),
		gMainTabGrid1, gMainTabLabel1);

	/*** scroll ***/
	gGeneralScroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_set_size_request (GTK_WIDGET (gGeneralScroll), 340, 0);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (gGeneralScroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	g_object_set (gGeneralScroll, "shadow-type", GTK_SHADOW_IN, NULL);
	gtk_grid_attach (GTK_GRID (gMainTabGrid1), gGeneralScroll, 0, 0, 1, 1);
	g_object_set (gGeneralScroll, "hexpand", FALSE, NULL);
	g_object_set (gGeneralScroll, "vexpand", TRUE, NULL);
	gtk_widget_show (gGeneralScroll);

	/*** grid ***/
	gGeneralGrid = gtk_grid_new();
	gtk_container_add (GTK_CONTAINER (gGeneralScroll), gGeneralGrid);
	gtk_widget_show (gGeneralGrid);

	/*** label ***/
	gGeneralLabel = gtk_label_new (NULL);
/***
Even if I use gtk_container_set_focus_chain() with NULL, the label
still keeps getting automatically selected fully now and then.
You would think, all right let's not make it selectable at all, but
then the next widget gets automatically selected, which means the
focus is down the page, and the scroll bar moves down.
***/
	gtk_label_set_selectable (GTK_LABEL (gGeneralLabel), TRUE);
	g_object_set (gGeneralLabel, "margin", 10, NULL);
	g_object_set (gGeneralLabel, "halign", GTK_ALIGN_START, NULL);
	g_object_set (gGeneralLabel, "valign", GTK_ALIGN_START, NULL);
	g_object_set (gGeneralLabel, "hexpand", FALSE, NULL);
	g_object_set (gGeneralLabel, "vexpand", FALSE, NULL);
	gtk_label_set_line_wrap (GTK_LABEL (gGeneralLabel), TRUE);
	gtk_label_set_line_wrap_mode (GTK_LABEL (gGeneralLabel),
		PANGO_WRAP_WORD_CHAR);
	gtk_label_set_justify (GTK_LABEL (gGeneralLabel), GTK_JUSTIFY_FILL);
	gtk_grid_attach (GTK_GRID (gGeneralGrid), gGeneralLabel, 0, 0, 1, 1);
	gtk_widget_show (gGeneralLabel);

	/*** link button ***/
	gGeneralLink = gtk_link_button_new_with_label
		("http://www.popot.org/custom_levels.php", "at popot.org");
	gtk_widget_set_margin_bottom (gGeneralLink, 10);
	g_object_set (gGeneralLink, "hexpand", TRUE, NULL);
	gtk_grid_attach (GTK_GRID (gGeneralGrid), gGeneralLink, 0, 1, 1, 1);
	gtk_widget_hide (gGeneralLink); /*** Yes, hide. ***/

	/*** "Shots" ***/
	gMainTabLabel2 = gtk_label_new ("Shots (0)");
	gtk_widget_show (gMainTabLabel2);
	gMainTabGrid2 = gtk_grid_new();
	gtk_widget_show (gMainTabGrid2);
	gtk_notebook_append_page (GTK_NOTEBOOK (gMainTab),
		gMainTabGrid2, gMainTabLabel2);

	/*** scroll ***/
	gShotsScroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_set_size_request (GTK_WIDGET (gShotsScroll), 340, 0);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (gShotsScroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	g_object_set (gShotsScroll, "shadow-type", GTK_SHADOW_IN, NULL);
	gtk_grid_attach (GTK_GRID (gMainTabGrid2), gShotsScroll, 0, 0, 1, 1);
	g_object_set (gShotsScroll, "hexpand", FALSE, NULL);
	g_object_set (gShotsScroll, "vexpand", TRUE, NULL);
	gtk_widget_show (gShotsScroll);

	/*** grid ***/
	gShotsGrid = gtk_grid_new();
	gtk_container_add (GTK_CONTAINER (gShotsScroll), gShotsGrid);
	gtk_widget_show (gShotsGrid);

	/*** Add default images. ***/
	for (iShotLoop = 1; iShotLoop <= 9; iShotLoop++)
	{
		snprintf (sPNG, MAX_LOCATION, "png%sshot.png", SLASH);
		gShotsImage[iShotLoop] = gtk_image_new_from_file (sPNG);
		gtk_grid_attach (GTK_GRID (gShotsGrid), gShotsImage[iShotLoop],
			0, iShotLoop - 1, 1, 1);
		g_object_set (gShotsImage[iShotLoop], "hexpand", TRUE, NULL);
		gtk_widget_hide (gShotsImage[iShotLoop]); /*** Yes, hide. ***/
	}

	/*** "Comments" ***/
	gMainTabLabel3 = gtk_label_new ("Comments (0)");
	gtk_widget_show (gMainTabLabel3);
	gMainTabGrid3 = gtk_grid_new();
	gtk_widget_show (gMainTabGrid3);
	gtk_notebook_append_page (GTK_NOTEBOOK (gMainTab),
		gMainTabGrid3, gMainTabLabel3);

	/*** scroll ***/
	gCommentsScroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_set_size_request (GTK_WIDGET (gCommentsScroll), 340, 0);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (gCommentsScroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	g_object_set (gCommentsScroll, "shadow-type", GTK_SHADOW_IN, NULL);
	gtk_grid_attach (GTK_GRID (gMainTabGrid3), gCommentsScroll, 0, 0, 1, 1);
	g_object_set (gCommentsScroll, "hexpand", FALSE, NULL);
	g_object_set (gCommentsScroll, "vexpand", TRUE, NULL);
	gtk_widget_show (gCommentsScroll);

	/*** label ***/
	gCommentsLabel = gtk_label_new (NULL);
	gtk_label_set_selectable (GTK_LABEL (gCommentsLabel), TRUE);
	g_object_set (gCommentsLabel, "margin", 10, NULL);
	g_object_set (gCommentsLabel, "halign", GTK_ALIGN_START, NULL);
	g_object_set (gCommentsLabel, "valign", GTK_ALIGN_START, NULL);
	g_object_set (gCommentsLabel, "hexpand", FALSE, NULL);
	g_object_set (gCommentsLabel, "vexpand", FALSE, NULL);
	gtk_label_set_line_wrap (GTK_LABEL (gCommentsLabel), TRUE);
	gtk_label_set_line_wrap_mode (GTK_LABEL (gCommentsLabel),
		PANGO_WRAP_WORD_CHAR);
	gtk_label_set_justify (GTK_LABEL (gCommentsLabel), GTK_JUSTIFY_FILL);
	gtk_container_add (GTK_CONTAINER (gCommentsScroll), gCommentsLabel);
	gtk_widget_show (gCommentsLabel);

	/*** "Replays" ***/
	gMainTabLabel4 = gtk_label_new ("Replays (0)");
	gtk_widget_show (gMainTabLabel4);
	gMainTabGrid4 = gtk_grid_new();
	gtk_widget_show (gMainTabGrid4);
	gtk_notebook_append_page (GTK_NOTEBOOK (gMainTab),
		gMainTabGrid4, gMainTabLabel4);

	/*** scroll ***/
	gReplaysScroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_set_size_request (GTK_WIDGET (gReplaysScroll), 340, 0);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (gReplaysScroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	g_object_set (gReplaysScroll, "shadow-type", GTK_SHADOW_IN, NULL);
	gtk_grid_attach (GTK_GRID (gMainTabGrid4), gReplaysScroll, 0, 0, 1, 1);
	g_object_set (gReplaysScroll, "hexpand", FALSE, NULL);
	g_object_set (gReplaysScroll, "vexpand", TRUE, NULL);
	gtk_widget_show (gReplaysScroll);

	/*** label ***/
	gReplaysLabel = gtk_label_new (NULL);
	gtk_label_set_use_markup (GTK_LABEL (gReplaysLabel), TRUE);
	gtk_label_set_selectable (GTK_LABEL (gReplaysLabel), TRUE);
	g_object_set (gReplaysLabel, "margin", 10, NULL);
	g_object_set (gReplaysLabel, "halign", GTK_ALIGN_START, NULL);
	g_object_set (gReplaysLabel, "valign", GTK_ALIGN_START, NULL);
	g_object_set (gReplaysLabel, "hexpand", FALSE, NULL);
	g_object_set (gReplaysLabel, "vexpand", FALSE, NULL);
	gtk_label_set_line_wrap (GTK_LABEL (gReplaysLabel), TRUE);
	gtk_label_set_line_wrap_mode (GTK_LABEL (gReplaysLabel),
		PANGO_WRAP_WORD_CHAR);
	gtk_label_set_justify (GTK_LABEL (gReplaysLabel), GTK_JUSTIFY_FILL);
	gtk_container_add (GTK_CONTAINER (gReplaysScroll), gReplaysLabel);
	gtk_widget_show (gReplaysLabel);

	gtk_notebook_set_current_page (GTK_NOTEBOOK (gMainTab), 0);

	/*** button ***/
	gMainButtonGet = gtk_button_new_with_label ("Get");
	gtk_widget_set_size_request (GTK_WIDGET (gMainButtonGet), 160, 0);
	g_signal_connect (gMainButtonGet, "clicked",
		G_CALLBACK (Get), NULL);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainButtonGet, 4, 3, 1, 1);
	gtk_widget_set_margin_start (gMainButtonGet, 10);
	gtk_widget_set_margin_bottom (gMainButtonGet, 10);
	g_object_set (gMainButtonGet, "sensitive", FALSE, NULL);
	gtk_widget_show (gMainButtonGet);

	/*** button ***/
	gMainButtonPlay = gtk_button_new_with_label ("Play");
	gtk_widget_set_size_request (GTK_WIDGET (gMainButtonPlay), 160, 0);
	g_signal_connect (gMainButtonPlay, "clicked",
		G_CALLBACK (Play), NULL);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainButtonPlay, 5, 3, 1, 1);
	gtk_widget_set_margin_end (gMainButtonPlay, 10);
	gtk_widget_set_margin_bottom (gMainButtonPlay, 10);
	g_object_set (gMainButtonPlay, "sensitive", FALSE, NULL);
	gtk_widget_show (gMainButtonPlay);

	/*** button ***/
	gMainCButton = gtk_check_button_new_with_label ("Enable cheats.");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (gMainCButton), FALSE);
	gtk_grid_attach (GTK_GRID (gMainGrid), gMainCButton, 4, 4, 1, 1);
	g_object_set (gMainCButton, "valign", GTK_ALIGN_CENTER, NULL);
	gtk_widget_set_margin_start (gMainCButton, 10);
	gtk_widget_set_margin_bottom (gMainCButton, 10);
	g_signal_connect (gMainCButton, "clicked",
		G_CALLBACK (Cheat), NULL);
	gtk_widget_show (gMainCButton);

	/*** grid ***/
	gLevelGrid = gtk_grid_new();
	gtk_grid_attach (GTK_GRID (gMainGrid), gLevelGrid, 5, 4, 1, 1);
	gtk_widget_set_halign (gLevelGrid, GTK_ALIGN_END);
	gtk_widget_set_margin_end (gLevelGrid, 10);
	gtk_widget_set_margin_bottom (gLevelGrid, 10);
	gtk_widget_show (gLevelGrid);

	/*** label ***/
	gLevelLabel = gtk_label_new ("Level:");
	gtk_grid_attach (GTK_GRID (gLevelGrid), gLevelLabel, 0, 0, 1, 1);
	g_object_set (gLevelLabel, "margin", 10, NULL);
	gtk_widget_set_margin_start (gLevelLabel, 0);
	gtk_widget_show (gLevelLabel);

	/*** combo box ***/
	gLevelCombo = gtk_combo_box_new();
	g_signal_connect (gLevelCombo, "changed",
		G_CALLBACK (Level), NULL);
	/***/
	InsertLevels();
	/***/
	gtk_grid_attach (GTK_GRID (gLevelGrid), gLevelCombo, 1, 0, 1, 1);
	gtk_combo_box_set_active (GTK_COMBO_BOX (gLevelCombo), 0);
	gtk_widget_show (gLevelCombo);

	gtk_widget_show (gMainWindow);
}
/*****************************************************************************/
void Quit (GtkWidget *widget, gpointer user_data)
/*****************************************************************************/
{
	/* GtkWidget *widget receives either GtkMenuItem *menuitem
	 * or GtkWidget *widget.
	 */

	if (widget != NULL) { }
	if (user_data != NULL) { }

	xmlCleanupParser();
	exit (EXIT_NORMAL);
}
/*****************************************************************************/
gboolean QuitEvent (GtkWidget *widget, GdkEvent *event, gpointer user_data)
/*****************************************************************************/
{
	if (event != NULL) { }

	Quit (widget, user_data);

	return (TRUE);
}
/*****************************************************************************/
void UpdateList (void)
/*****************************************************************************/
{
	GtkTreeIter iter;
	int iModLoop;
	GdkPixbuf *gIcon;
	char sType[20 + 2];

	gtk_list_store_clear (gMainListStore);
	for (iModLoop = 1; iModLoop <= iNrModsNew; iModLoop++)
	{
		gtk_list_store_append (gMainListStore, &iter);
		gtk_tree_model_iter_nth_child (gMainTreeModel, &iter, NULL, iModLoop - 1);
		gtk_list_store_set (gMainListStore, &iter, 0,
			arMods_S_Number[iModLoop], -1);
		switch (arMods_I_PoP_Version[iModLoop])
		{
			case 1: /*** PoP1 for DOS, SDLPoP, MININIM ***/
				if (strcmp (arMods_S_Executable_DOSSNES[iModLoop], "") != 0)
				{
					snprintf (sType, 20, "%s", "PoP1_DOS");
				} else if (strcmp (arMods_S_Executable_SDLPoP[iModLoop], "") != 0) {
					snprintf (sType, 20, "%s", "PoP1_SDLPoP");
				} else if (strcmp (arMods_S_Executable_MININIM[iModLoop], "") != 0) {
					snprintf (sType, 20, "%s", "PoP1_MININIM");
				}
				break;
			case 2: /*** PoP2 for DOS ***/
				snprintf (sType, 20, "%s", "PoP2_DOS");
				break;
			case 4: /*** PoP1 for SNES ***/
				snprintf (sType, 20, "%s", "PoP1_SNES");
				break;
		}
		snprintf (sPNG, MAX_LOCATION, "png%s%s.png", SLASH, sType);
		gIcon = gdk_pixbuf_new_from_file (sPNG, NULL);
		gtk_list_store_set (gMainListStore, &iter, 1,
			gIcon, -1);
		gtk_list_store_set (gMainListStore, &iter, 2,
			arMods_S_Name[iModLoop], -1);
	}
}
/*****************************************************************************/
void UpdateShots (void)
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iShotLoop;

	for (iShotLoop = 1; iShotLoop <= 9; iShotLoop++)
	{
		if (iShotLoop <= arMods_I_Screenshots[iModShowing])
		{
			snprintf (sPNG, MAX_LOCATION, "%s%s%s_%i.png",
				MODS_DIR, SLASH, arMods_S_Number[iModShowing], iShotLoop);
			if (access (sPNG, R_OK) == -1)
			{
				snprintf (sPNG, MAX_LOCATION, "png%sshot.png", SLASH);
			}
			gtk_image_set_from_file (GTK_IMAGE (gShotsImage[iShotLoop]), sPNG);
			gtk_widget_show (gShotsImage[iShotLoop]);
		} else {
			gtk_widget_hide (gShotsImage[iShotLoop]);
		}
	}
}
/*****************************************************************************/
void Get (GtkButton *button, gpointer user_data)
/*****************************************************************************/
{
	if (button != NULL) { }
	if (user_data != NULL) { }

	iWait = 1 + arMods_I_Screenshots[iModShowing];
	g_object_set (gMainWindow, "sensitive", FALSE, NULL);
	PleaseWait();

	GetMod (arMods_S_Number[iModShowing]);
	GetShots (arMods_S_Number[iModShowing]);

	do {
		gtk_main_iteration();
	} while (iWait > 0);
	g_usleep (G_USEC_PER_SEC); /*** One extra second, just in case. ***/
	g_object_set (gMainWindow, "sensitive", TRUE, NULL);
	gtk_widget_destroy (gWaitWindow);
	gWaitWindow = NULL;

	UnzipMod (arMods_S_Number[iModShowing]);
	snprintf (sTXT, MAX_LOCATION, "%s%s%s.TXT",
		MODS_DIR, SLASH, arMods_S_Number[iModShowing]);
	remove (sTXT);
	UpdateButtons();
	UpdateShots();
}
/*****************************************************************************/
void Play (GtkButton *button, gpointer user_data)
/*****************************************************************************/
{
	if (button != NULL) { }
	if (user_data != NULL) { }

	CreateBatchFile (arMods_S_Number[iModShowing]);
	RunGame (arMods_S_Number[iModShowing]);
}
/*****************************************************************************/
void Cheat (GtkButton *button, gpointer user_data)
/*****************************************************************************/
{
	if (button != NULL) { }
	if (user_data != NULL) { }

	if (iModCheat == 0)
	{
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (gMainCButton), TRUE);
		iModCheat = 1;
	} else {
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (gMainCButton), FALSE);
		iModCheat = 0;

		/*** No specific level. ***/
		gtk_combo_box_set_active (GTK_COMBO_BOX (gLevelCombo), 0);
	}
}
/*****************************************************************************/
void SelectRow (GtkTreeSelection *gSelection, gpointer user_data)
/*****************************************************************************/
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreePath *path;
	/***/
	gint *indices;
	gchar *value;

	if (user_data != NULL) { }

	if (gtk_tree_selection_get_selected (
		GTK_TREE_SELECTION (gSelection), &model, &iter))
	{
		/*** row ***/
		path = gtk_tree_model_get_path (model, &iter);
		indices = gtk_tree_path_get_indices (path);
		iModActive = indices[0] + 1;

		/*** number ***/
		gtk_tree_model_get (model, &iter, 0, &value, -1);
		if (iDebug == 1)
		{
			printf ("[ INFO ] Selected mod: %s\n", value);
		}
		g_free (value);
	}

	/* Only change the buttons and show the mod if it's not
	 * already being shown.
	 */
	if (iModActive == iModShowing) { return; }
	iModShowing = iModActive;

	UpdateButtons();

	ShowMod();
}
/*****************************************************************************/
void ShowMod (void)
/*****************************************************************************/
{
	char sText[TEXT_HUGE + 2];
	char sTextTemp[TEXT_HUGE + 2];
	char sAuthors[200 + 2];
	char sAuthorsTemp[200 + 2];
	char sYear[10 + 2];
	char sMinutes[10 + 2];
	char sCGraphics[10 + 2];
	char sCAudio[10 + 2];
	char sEnjoyment[10 + 2];
	char sDifficulty[10 + 2];
	char sLabel[100 + 2];
	char sURL[60 + 2];
	char sHref[200 + 2];
	char sComment[TEXT_LARGE + 2];

	/*** Used for looping. ***/
	int iCommentsLoop;
	int iReplaysLoop;

	/*** Authors. ***/
	snprintf (sAuthors, 200, "%s", arMods_S_Author1_Name[iModShowing]);
	if ((strcmp (arMods_S_Author2_Name[iModShowing], "") != 0) &&
		(strcmp (arMods_S_Author2_Name[iModShowing], "?") != 0))
	{
		snprintf (sAuthorsTemp, 200, "%s", sAuthors);
		snprintf (sAuthors, 200, "%s, %s", sAuthorsTemp,
			arMods_S_Author2_Name[iModShowing]);
	}
	if ((strcmp (arMods_S_Author3_Name[iModShowing], "") != 0) &&
		(strcmp (arMods_S_Author3_Name[iModShowing], "?") != 0))
	{
		snprintf (sAuthorsTemp, 200, "%s", sAuthors);
		snprintf (sAuthors, 200, "%s, %s", sAuthorsTemp,
			arMods_S_Author3_Name[iModShowing]);
	}

	/*** Year. ***/
	snprintf (sYear, 10, "%i", arMods_I_Creation_Year[iModShowing]);
	if (strcmp (sYear, "0") == 0)
		{ snprintf (sYear, 10, "%s", "?"); }

	/*** Minutes. ***/
	snprintf (sMinutes, 10, "%i", arMods_I_Starting_Minutes[iModShowing]);
	if (strcmp (sMinutes, "0") == 0)
		{ snprintf (sMinutes, 10, "%s", "?"); }

	/*** Graphics. ***/
	switch (arMods_I_Altered_Graphics[iModShowing])
	{
		case 0: snprintf (sCGraphics, 10, "%s", "no"); break;
		case 1: snprintf (sCGraphics, 10, "%s", "yes"); break;
	}

	/*** Audio. ***/
	switch (arMods_I_Altered_Audio[iModShowing])
	{
		case 0: snprintf (sCAudio, 10, "%s", "no"); break;
		case 1: snprintf (sCAudio, 10, "%s", "yes"); break;
	}

	/*** Enjoyment. ***/
	if (arMods_I_Enjoyment_Votes[iModShowing] == 0)
	{
		snprintf (sEnjoyment, 10, "%s", "-");
	} else {
		snprintf (sEnjoyment, 10, "%.2f", arMods_F_Enjoyment[iModShowing]);
	}

	/*** Difficulty. ***/
	if (arMods_I_Difficulty_Votes[iModShowing] == 0)
	{
		snprintf (sDifficulty, 10, "%s", "-");
	} else {
		snprintf (sDifficulty, 10, "%.2f", arMods_F_Difficulty[iModShowing]);
	}

	gtk_label_set_text (GTK_LABEL (gMainLabel2), arMods_S_Name[iModShowing]);

	/*** Update the tab labels. ***/
	snprintf (sLabel, 100, "Shots (%i)",
		arMods_I_Screenshots[iModShowing]);
	gtk_label_set_text (GTK_LABEL (gMainTabLabel2), sLabel);
	snprintf (sLabel, 100, "Comments (%i)",
		arMods_I_Comments[iModShowing]);
	gtk_label_set_text (GTK_LABEL (gMainTabLabel3), sLabel);
	snprintf (sLabel, 100, "Replays (%i)",
		arMods_I_Replays[iModShowing]);
	gtk_label_set_text (GTK_LABEL (gMainTabLabel4), sLabel);

	snprintf (sText, TEXT_HUGE,
		"author(s): %s\n" \
		"year: %s\n" \
		"minutes: %s\n" \
		"custom levels: %i\n" \
		"custom graphics: %s\n" \
		"custom audio: %s\n" \
		"enjoyment: %s (%i votes)\n" \
		"difficulty: %s (%i votes)\n" \
		"last update: %s\n" \
		"last ZIP update: %s\n" \
		"\ndescription:\n%s",
		sAuthors, sYear, sMinutes,
		arMods_I_Levels_Changed[iModShowing],
		sCGraphics, sCAudio,
		sEnjoyment, arMods_I_Enjoyment_Votes[iModShowing],
		sDifficulty, arMods_I_Difficulty_Votes[iModShowing],
		arMods_S_Last_Update[iModShowing],
		arMods_S_Last_Update_ZIP[iModShowing],
		arMods_S_Description[iModShowing]);
	gtk_label_set_text (GTK_LABEL (gGeneralLabel), sText);

	snprintf (sURL, 60, "http://www.popot.org/custom_levels.php?mod=%s",
		arMods_S_Number[iModShowing]);
	gtk_link_button_set_uri (GTK_LINK_BUTTON (gGeneralLink), sURL);
	gtk_widget_show (gGeneralLink);

	UpdateShots();

	snprintf (sText, TEXT_HUGE, "%s", "");
	for (iCommentsLoop = 1; iCommentsLoop <=
		arMods_I_Comments[iModShowing]; iCommentsLoop++)
	{
		if (strcmp (sText, "") == 0)
		{
			snprintf (sTextTemp, TEXT_HUGE, "%s", sText);
		} else {
			snprintf (sTextTemp, TEXT_HUGE, "%s\n\n", sText);
		}
		snprintf (sText, TEXT_HUGE, "%s%s (%s):\n%s",
			sTextTemp,
			arComments_S_Nick[iModShowing][iCommentsLoop],
			arComments_S_Added[iModShowing][iCommentsLoop],
			arComments_S_Comment[iModShowing][iCommentsLoop]);
	}
	gtk_label_set_text (GTK_LABEL (gCommentsLabel), sText);

	snprintf (sText, TEXT_HUGE, "%s", "");
	for (iReplaysLoop = 1; iReplaysLoop <=
		arMods_I_Replays[iModShowing]; iReplaysLoop++)
	{
		if (strcmp (sText, "") == 0)
		{
			snprintf (sTextTemp, TEXT_HUGE, "%s", sText);
		} else {
			snprintf (sTextTemp, TEXT_HUGE, "%s\n\n", sText);
		}

		/*** Href. ***/
		snprintf (sHref, 200, "<a href=\"%s\">download</a>",
			arReplays_S_URL[iModShowing][iReplaysLoop]);

		/*** Comment. ***/
		if (strcmp (arReplays_S_Comment[iModShowing][iReplaysLoop], "") == 0)
		{
			snprintf (sComment, TEXT_LARGE, "%s", "");
		} else {
			snprintf (sComment, TEXT_LARGE, "; \"%s\"",
				arReplays_S_Comment[iModShowing][iReplaysLoop]);
		}

		snprintf (sText, TEXT_HUGE, "%s%s (%s): %s\n(%s; %s%s)",
			sTextTemp,
			arReplays_S_Level[iModShowing][iReplaysLoop],
			arReplays_S_Program[iModShowing][iReplaysLoop],
			sHref,
			arReplays_S_Nick[iModShowing][iReplaysLoop],
			arReplays_S_Added[iModShowing][iReplaysLoop],
			sComment);
	}
	gtk_label_set_markup (GTK_LABEL (gReplaysLabel), sText);

	if (arMods_I_PoP_Version[iModShowing] == 4) /*** PoP1 for SNES ***/
	{
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (gMainCButton), FALSE);
		iModCheat = 0;
		g_object_set (gMainCButton, "sensitive", FALSE, NULL);
		gtk_combo_box_set_active (GTK_COMBO_BOX (gLevelCombo), 0);
		g_object_set (gLevelCombo, "sensitive", FALSE, NULL);
	} else {
		g_object_set (gMainCButton, "sensitive", TRUE, NULL);
		g_object_set (gLevelCombo, "sensitive", TRUE, NULL);
	}
}
/*****************************************************************************/
void GetMenu (GtkWidget **widget, int iMenu, int iX)
/*****************************************************************************/
{
	GtkWidget *menu_file;
	GtkWidget *menu_help;
	GtkWidget *menuitemmain_file;
	GtkWidget *menuitemmain_help;

	*widget = gtk_menu_bar_new();

	if (iMenu == 1)
	{
		menu_file = gtk_menu_new();
		menuitemmain_file = gtk_menu_item_new_with_mnemonic ("_File");
		gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitemmain_file), menu_file);
		AddMenuItem (&menu_file, "_Quit", "application-exit", "Quit", iX);
		gtk_menu_shell_append (GTK_MENU_SHELL (*widget), menuitemmain_file);

		/***/

		menu_help = gtk_menu_new();
		gtk_widget_set_hexpand (menu_help, TRUE);
		gtk_widget_set_halign (menu_help, GTK_ALIGN_FILL);
		menuitemmain_help = gtk_menu_item_new_with_mnemonic ("_Help");
		g_object_set (menuitemmain_help, "right_justified", TRUE, NULL);
		gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitemmain_help), menu_help);
		AddMenuItem (&menu_help, "A_bout...", "help-about", "About", iX);
		gtk_menu_shell_append (GTK_MENU_SHELL (*widget), menuitemmain_help);
	}
}
/*****************************************************************************/
void AddMenuItem (GtkWidget **gMenu, char *sEntry,
	char *sIcon, char *sFunc, int iX)
/*****************************************************************************/
{
	GtkWidget *menuitem;
	GtkWidget *menuimage;

/* GTK+ more or less deprecated images on menu items, but we
 * still want to use them. We don't want warnings about it.
 * (Use gtk3-icon-browser to get an overview of GTK+ icons.)
 */
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	menuitem = gtk_image_menu_item_new_with_mnemonic (sEntry);
	menuimage = gtk_image_new_from_icon_name (sIcon,
		GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		GTK_WIDGET (menuimage));
#pragma GCC diagnostic pop
	gtk_menu_shell_append (GTK_MENU_SHELL (*gMenu), menuitem);

	if (strcmp (sFunc, "Quit") == 0)
	{
		g_signal_connect (menuitem, "activate",
			G_CALLBACK (Quit), GINT_TO_POINTER (iX));
	}
	if (strcmp (sFunc, "About") == 0)
	{
		g_signal_connect (menuitem, "activate",
			G_CALLBACK (About), GINT_TO_POINTER (iX));
	}
}
/*****************************************************************************/
void About (GtkMenuItem *menuitem, gpointer user_data)
/*****************************************************************************/
{
	GtkWidget *aboutdialog;
	GdkPixbuf *abouticon;

	if (menuitem != NULL) { }
	if (user_data != NULL) { }

	/*** dialog ***/
	aboutdialog = gtk_about_dialog_new();
	gtk_window_set_transient_for (GTK_WINDOW (aboutdialog),
		GTK_WINDOW (gMainWindow));
	gtk_about_dialog_set_program_name (GTK_ABOUT_DIALOG (aboutdialog),
		PROG_NAME);
	gtk_about_dialog_set_version (GTK_ABOUT_DIALOG (aboutdialog),
		"v" PROG_VERSION " (" PROG_MONTH_YEAR ")");
	gtk_about_dialog_set_copyright (GTK_ABOUT_DIALOG (aboutdialog),
		PROG_COPY_FULL);
	gtk_about_dialog_set_comments (GTK_ABOUT_DIALOG (aboutdialog),
		PROG_DESC);
	gtk_about_dialog_set_license_type (GTK_ABOUT_DIALOG (aboutdialog),
		GTK_LICENSE_GPL_3_0);
	gtk_about_dialog_set_website (GTK_ABOUT_DIALOG (aboutdialog),
		PROG_WEBSITE);
	gtk_about_dialog_set_website_label (GTK_ABOUT_DIALOG (aboutdialog),
		PROG_NAME " website");
	const gchar *authors[] = { PROG_AUTHOR_MAIN, NULL };
	gtk_about_dialog_set_authors (GTK_ABOUT_DIALOG (aboutdialog), authors);
	snprintf (sPNG, MAX_LOCATION, "png%s%s%s", SLASH, PROG_NAME, "_icon.png");
	abouticon = gdk_pixbuf_new_from_file (sPNG, NULL);
	gtk_about_dialog_set_logo (GTK_ABOUT_DIALOG (aboutdialog), abouticon);
	gtk_window_set_title (GTK_WINDOW (aboutdialog), "About");
	gtk_dialog_run (GTK_DIALOG (aboutdialog));

	/*** Always remove the dialog. ***/
	gtk_widget_destroy (aboutdialog);
}
/*****************************************************************************/
void ShowUsage (void)
/*****************************************************************************/
{
	printf ("%s v%s (%s)\n%s\n\n",
		PROG_NAME, PROG_VERSION, PROG_MONTH_YEAR, PROG_COPY_LAST);
	printf ("Usage:\n");
	printf ("  %s [OPTIONS]\n\nOptions:\n", PROG_NAME);
	printf ("  -h, -?,    --help           display this help and exit\n");
	printf ("  -v,        --version        output version information and"
		" exit\n");
	printf ("  -d,        --debug          show debug information\n");
	printf ("\n");

	exit (EXIT_NORMAL);
}
/*****************************************************************************/
gboolean GlobalSearch (GtkTreeModel *model, gint column,
	const gchar *key, GtkTreeIter *iter, gpointer search_data)
/*****************************************************************************/
{
	/*** Returns FALSE [sic] if there is a match. ***/

	gchar *iter2;
	gchar *found;

	if (search_data != NULL) { }

	gtk_tree_model_get (model, iter, column, &iter2, -1);
	found = g_strrstr (g_utf8_strdown (iter2, -1), g_utf8_strdown (key, -1));

	if (found == NULL) { return TRUE; } else { return FALSE; }
}
/*****************************************************************************/
void Refresh (GtkButton *button, gpointer user_data)
/*****************************************************************************/
{
	if (button != NULL) { }
	if (user_data != NULL) { }

	gtk_entry_set_text (GTK_ENTRY (gMainEntry), "");
	iModActive = 0;
	ParseXML();
	UpdateList();
}
/*****************************************************************************/
void ParseXML (void)
/*****************************************************************************/
{
	char sLocation[MAX_LOCATION + 2];
	char sModDateOld[DATE + 2];
	char sModDateOldZIP[DATE + 2];
	FILE *fTXT;
	char sWrite[MAX_WRITE + 2];

	/*** Used for looping. ***/
	int iModLoop;
	int iOldModLoop;

	snprintf (sLocation, MAX_LOCATION, "%s%s%s", XML_DIR, SLASH, XML_MODSU);

	/*** Get old date. ***/
	if (access (sLocation, R_OK) == -1)
	{
		snprintf (sModsUDateOld, DATE, "%s", "1970-01-01 00:00:00");
	} else {
		iDateWhen = 0;
		ModsU (sLocation);
	}

	/*** Get new date. ***/
	GetXML (XML_MODSU);
	iDateWhen = 1;
	ModsU (sLocation);

	snprintf (sLocation, MAX_LOCATION, "%s%s%s", XML_DIR, SLASH, XML_MODS);

	/*** Get old modification dates. ***/
	iNrModsOld = 0;
	if (access (sLocation, R_OK) != -1)
	{
		Mods (sLocation, 0);
	}

	/*** Get all data. ***/
	iNrModsNew = 0;
	if ((access (sLocation, R_OK) == -1) ||
		(DateTimeToPosix (sModsUDateOld) < DateTimeToPosix (sModsUDateNew)))
	{ GetXML (XML_MODS); }
	Mods (sLocation, 1);

	/*** Has the mod been updated? ***/
	for (iModLoop = 1; iModLoop < iNrModsNew; iModLoop++)
	{
		snprintf (sModDateOld, DATE, "%s", "1970-01-01 00:00:00");
		snprintf (sModDateOldZIP, DATE, "%s", "1970-01-01 00:00:00");
		for (iOldModLoop = 1; iOldModLoop < iNrModsOld; iOldModLoop++)
		{
			if (strcmp (arOldMods_S_Number[iOldModLoop],
				arMods_S_Number[iModLoop]) == 0)
			{
				snprintf (sModDateOld, DATE, "%s",
					arOldMods_S_Last_Update[iOldModLoop]);
				snprintf (sModDateOldZIP, DATE, "%s",
					arOldMods_S_Last_Update_ZIP[iOldModLoop]);
			}
		}
		if (DateTimeToPosix (sModDateOld) <
			DateTimeToPosix (arMods_S_Last_Update[iModLoop]))
		{
			if (iDebug == 1)
			{
				printf ("[ INFO ] Mod \"%s\" has been updated.\n",
					arMods_S_Number[iModLoop]);
			}
		}
		snprintf (sModDir, MAX_LOCATION, "%s%s%s", MODS_DIR, SLASH,
			arMods_S_Number[iModLoop]);
		if ((DateTimeToPosix (sModDateOldZIP) <
			DateTimeToPosix (arMods_S_Last_Update_ZIP[iModLoop])) ||
			(stat (sModDir, &stStatus) == -1))
		{
			if (iDebug == 1)
			{
				printf ("[ INFO ] Mod \"%s\" needs to be (re)downloaded.\n",
					arMods_S_Number[iModLoop]);
			}
			snprintf (sTXT, MAX_LOCATION, "%s%s%s.TXT",
				MODS_DIR, SLASH, arMods_S_Number[iModLoop]);
			fTXT = fopen (sTXT, "w");
			snprintf (sWrite, MAX_WRITE, "%s",
				"Please do not delete this file.\n" \
				"This mod needs to be (re)downloaded.\n");
			fwrite (sWrite, 1, strlen (sWrite), fTXT);
			fclose (fTXT);
		}
	}
}
/*****************************************************************************/
void UpdateButtons (void)
/*****************************************************************************/
{
	gboolean bExistsTXT, bExistsDir;

	/*** Does this mod have a TXT or directory? ***/
	snprintf (sTXT, MAX_LOCATION, "%s%s%s.TXT",
		MODS_DIR, SLASH, arMods_S_Number[iModActive]);
	if (access (sTXT, R_OK) == -1)
		{ bExistsTXT = FALSE; } else { bExistsTXT = TRUE; }
	snprintf (sModDir, MAX_LOCATION, "%s%s%s", MODS_DIR, SLASH,
		arMods_S_Number[iModActive]);
	if (stat (sModDir, &stStatus) == -1)
		{ bExistsDir = FALSE; } else { bExistsDir = TRUE; }

	/*** Get / Update ***/
	if (bExistsDir == TRUE)
	{
		if (bExistsTXT == TRUE)
		{
			g_object_set (gMainButtonGet, "sensitive", TRUE, NULL);
			gtk_button_set_label (GTK_BUTTON (gMainButtonGet), "Update");
		} else {
			g_object_set (gMainButtonGet, "sensitive", FALSE, NULL);
			gtk_button_set_label (GTK_BUTTON (gMainButtonGet), "Update");
		}
	} else {
		g_object_set (gMainButtonGet, "sensitive", TRUE, NULL);
		gtk_button_set_label (GTK_BUTTON (gMainButtonGet), "Get");
	}

	/*** Play ***/
	g_object_set (gMainButtonPlay, "sensitive", bExistsDir, NULL);
}
/*****************************************************************************/
void GetMod (char *sNumber)
/*****************************************************************************/
{
	char sFrom[MAX_LOCATION + 2], sTo[MAX_LOCATION + 2];
	GFile *remote, *local;

	snprintf (sFrom, MAX_LOCATION,
		"http://www.popot.org/custom_levels/software/%s.zip", sNumber);
	remote = g_file_new_for_uri (sFrom);
	snprintf (sTo, MAX_LOCATION, "%s%s%s.zip", MODS_DIR, SLASH, sNumber);
	local = g_file_new_for_path (sTo);

	if (iDebug == 1)
	{
		printf ("[ INFO ] Copying \"%s\" to \"%s\".\n", sFrom, sTo);
	}

	g_file_copy_async (remote, local, G_FILE_COPY_OVERWRITE,
		G_PRIORITY_DEFAULT, NULL, NULL, NULL, Done, NULL);
}
/*****************************************************************************/
void GetShots (char *sNumber)
/*****************************************************************************/
{
	char sFrom[MAX_LOCATION + 2], sTo[MAX_LOCATION + 2];
	GFile *remote, *local;

	/*** Used for looping. ***/
	int iShotLoop;

	for (iShotLoop = 1; iShotLoop <=
		arMods_I_Screenshots[iModShowing]; iShotLoop++)
	{
		snprintf (sFrom, MAX_LOCATION,
			"http://www.popot.org/custom_levels/screenshots/%s_%i.png",
			sNumber, iShotLoop);
		remote = g_file_new_for_uri (sFrom);
		snprintf (sTo, MAX_LOCATION, "%s%s%s_%i.png",
			MODS_DIR, SLASH, sNumber, iShotLoop);
		local = g_file_new_for_path (sTo);

		if (iDebug == 1)
		{
			printf ("[ INFO ] Copying \"%s\" to \"%s\".\n", sFrom, sTo);
		}

		g_file_copy_async (remote, local, G_FILE_COPY_OVERWRITE,
			G_PRIORITY_DEFAULT, NULL, NULL, NULL, Done, NULL);
	}
}
/*****************************************************************************/
void UnzipMod (char *sNumber)
/*****************************************************************************/
{
	struct zip *zip;
	struct zip_stat zips;
	struct zip_file *zipf;
	/***/
	char sLocation[MAX_LOCATION + 2];
	char sZIP[MAX_LOCATION + 2];
	char sBuffer[100];
	unsigned long ulTotal;
	int iNrFiles;
	int iLength;
	int iError;
	int iChars;
	int iFd;

	/*** Used for looping. ***/
	int iFileLoop;

	snprintf (sZIP, MAX_LOCATION, "%s%s%s.zip", MODS_DIR, SLASH, sNumber);

	zip = zip_open (sZIP, 0, &iError);
	if (zip == NULL)
	{
		printf ("[ WARN ] Cannot open \"%s\": %s!\n",
			sZIP, strerror (errno));
	} else {
		iNrFiles = zip_get_num_entries (zip, 0);
		for (iFileLoop = 0; iFileLoop < iNrFiles; iFileLoop++)
		{
			if (zip_stat_index (zip, iFileLoop, 0, &zips) == 0)
			{
				snprintf (sLocation, MAX_LOCATION, "%s%s%s",
					MODS_DIR, SLASH, zips.name);
				iLength = strlen (zips.name);
				if (zips.name[iLength - 1] == '/')
				{
					CreateDir (sLocation);
				} else {
					zipf = zip_fopen_index (zip, iFileLoop, 0);
					iFd = open (sLocation, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0600);
					ulTotal = 0;
					while (ulTotal != zips.size)
					{
						iChars = zip_fread (zipf, sBuffer, 100);
						write (iFd, sBuffer, iChars);
						ulTotal+=iChars;
					}
					close (iFd);
					zip_fclose (zipf);
				}
			}
		}

		zip_close (zip);
	}
}
/*****************************************************************************/
void InsertLevels (void)
/*****************************************************************************/
{
	GtkTreeIter iter;
	GtkCellRenderer *renderer;
	GtkListStore *store;
	/***/
	char sLevel[10 + 2];

	/*** Used for looping. ***/
	int iLevelLoop;

	/*** renderer ***/
	renderer = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (gLevelCombo), renderer, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (gLevelCombo),
		renderer, "text", 0, NULL);

	/*** store ***/
	store = gtk_list_store_new (1, G_TYPE_STRING);
	for (iLevelLoop = 0; iLevelLoop <= 30; iLevelLoop++)
	{
		gtk_list_store_append (store, &iter);
		if (iLevelLoop == 0)
		{
			snprintf (sLevel, 10, "%s", NO_SPECIFIC_LEVEL);
		} else {
			snprintf (sLevel, 10, "%i", iLevelLoop);
		}
		gtk_list_store_set (store, &iter, 0, sLevel, -1);
	}

	/*** model ***/
	gLevelModel = GTK_TREE_MODEL (store);
	gtk_combo_box_set_model (GTK_COMBO_BOX (gLevelCombo), gLevelModel);
}
/*****************************************************************************/
void Level (GtkComboBox *widget, gpointer user_data)
/*****************************************************************************/
{
	GtkTreeIter iter;
	/***/
	gchar *value;

	if (user_data != NULL) { }

	/*** Occasionally, iter is NOT set, so if() is required. ***/
	if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter))
	{
		gtk_tree_model_get (gLevelModel, &iter, 0, &value, -1);
		if (strcmp (value, NO_SPECIFIC_LEVEL) == 0)
		{
			iModLevel = 0;
		} else {
			iModLevel = atoi (value);

			/*** Enable cheats. ***/
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (gMainCButton), TRUE);
			iModCheat = 1;
		}
	}
}
/*****************************************************************************/
void PleaseWait (void)
/*****************************************************************************/
{
	GtkWidget *gWaitLabel;

	/*** window ***/
	gWaitWindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_transient_for (GTK_WINDOW (gWaitWindow),
		GTK_WINDOW (gMainWindow));
	gtk_window_set_decorated (GTK_WINDOW (gWaitWindow), FALSE);
	gtk_window_set_position (GTK_WINDOW (gWaitWindow),
		GTK_WIN_POS_CENTER_ALWAYS);
	gtk_window_set_resizable (GTK_WINDOW (gWaitWindow), FALSE);
	gtk_widget_realize (gWaitWindow);

	/*** label ***/
	gWaitLabel = gtk_label_new ("Please wait...");
	gtk_label_set_justify (GTK_LABEL (gWaitLabel), GTK_JUSTIFY_CENTER);
	g_object_set (gWaitLabel, "margin", 10, NULL);
	gtk_container_add (GTK_CONTAINER (gWaitWindow), gWaitLabel);
	gtk_widget_show (gWaitLabel);

	gtk_widget_show (gWaitWindow);
}
/*****************************************************************************/
void Done (GObject *source_object, GAsyncResult *res, gpointer user_data)
/*****************************************************************************/
{
	GtkWidget *gDialog;
	GError *error = NULL; /*** Do NOT remove "= NULL;". ***/
	/***/
	gboolean bFinished;

	if (user_data != NULL) { }

	bFinished = g_file_copy_finish (G_FILE (source_object), res, &error);

	if ((bFinished == FALSE) || (error != NULL))
	{
		gDialog = gtk_message_dialog_new (GTK_WINDOW (gMainWindow),
			GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, "Could not download!");
		gtk_dialog_run (GTK_DIALOG (gDialog));
		gtk_widget_destroy (gDialog);
	}

	iWait--;
}
/*****************************************************************************/
void CreateBatchFile (char *sNumber)
/*****************************************************************************/
{
	/* Creates a batch file or shell script. This way we can enable
	 * cheats and (also) start in specific levels.
	 */

	int iFd;
	char sLocation[MAX_LOCATION + 2];
	char sToWrite[MAX_WRITE + 2];
	char sModLevel[10 + 2];
	char sModGameplay[10 + 2];
	char sModCheat[100 + 2];
	char sModEXE[100 + 2];

	if (arMods_I_PoP_Version[iModShowing] != 4) /*** Not for SNES mods. ***/
	{
		/*** Cheat. ***/
		snprintf (sModCheat, 100, "%s", "");
		if (iModCheat != 0)
		{
			snprintf (sModCheat, 100, " %s", arMods_S_Cheat_Code[iModShowing]);
		}

		/*** Level and gameplay. ***/
		snprintf (sModLevel, 10, "%s", "");
		snprintf (sModGameplay, 10, "%s", "");
		if (iModLevel != 0)
		{
			if (arMods_I_PoP_Version[iModShowing] == 2)
			{
				if (iModLevel <= 14)
				{
					snprintf (sModLevel, 10, " level%i", iModLevel);
				} else {
					snprintf (sModLevel, 10, " level%i", iModLevel - 14);
					snprintf (sModGameplay, 10, "%s", " gameplay");
				}
			} else {
				snprintf (sModLevel, 10, " %i", iModLevel);
			}
		}

		iFd = 0; /*** To prevent uninitialized warning. ***/

		/*** PoP1/PoP2 for DOS. ***/
		if (strcmp (arMods_S_Executable_DOSSNES[iModShowing], "") != 0)
		{
			snprintf (sLocation, MAX_LOCATION, "%s%s%s%s%s.bat",
				MODS_DIR, SLASH, sNumber, SLASH, PROG_NAME);
			iFd = open (sLocation, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0600);
			if (iFd == -1)
			{
				printf ("[ WARN ] CreateBatchFile(): %s!\n", strerror (errno));
			} else {
				snprintf (sToWrite, MAX_WRITE, "%s%s%s%s\n",
					arMods_S_Executable_DOSSNES[iModShowing],
					sModGameplay, sModCheat, sModLevel);
			}
		}

		/*** SDLPoP ***/
		if (strcmp (arMods_S_Executable_SDLPoP[iModShowing], "") != 0)
		{
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
snprintf (sLocation, MAX_LOCATION, "%s%s%s%s%s_s.bat", /*** _s.bat ***/
	MODS_DIR, SLASH, sNumber, SLASH, PROG_NAME);
iFd = open (sLocation, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0600); /*** 6 ***/
snprintf (sModEXE, 100, "%s", arMods_S_Executable_SDLPoP[iModShowing]);
#else
snprintf (sLocation, MAX_LOCATION, "%s%s%s%s%s_s.sh", /*** _s.sh ***/
	MODS_DIR, SLASH, sNumber, SLASH, PROG_NAME);
iFd = open (sLocation, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0700); /*** 7 ***/
snprintf (sModEXE, 100, "%s", arMods_S_Executable_SDLPoP[iModShowing]);
sModEXE[strlen (sModEXE) - 4] = '\0'; /*** Remove ".exe" ***/

/*** TODO: Maybe use zip_file_extra_field_get() instead? ***/
snprintf (sToWrite, MAX_WRITE, "chmod +x %s\n", sModEXE);
write (iFd, sToWrite, strlen (sToWrite));
#endif
			if (iFd == -1)
			{
				printf ("[ WARN ] CreateBatchFile(): %s!\n", strerror (errno));
			} else {
				snprintf (sToWrite, MAX_WRITE, "%s%s%s%s\n",
					HERE, sModEXE, sModCheat, sModLevel);
			}
		}

		/*** MININIM ***/
		if (strcmp (arMods_S_Executable_MININIM[iModShowing], "") != 0)
		{
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
snprintf (sLocation, MAX_LOCATION, "%s%s%s%s%s_m.bat", /*** _m.bat ***/
	MODS_DIR, SLASH, sNumber, SLASH, PROG_NAME);
iFd = open (sLocation, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0600); /*** 6 ***/
snprintf (sModEXE, 100, "%s", arMods_S_Executable_MININIM[iModShowing]);
#else
snprintf (sLocation, MAX_LOCATION, "%s%s%s%s%s_m.sh", /*** _m.sh ***/
	MODS_DIR, SLASH, sNumber, SLASH, PROG_NAME);
iFd = open (sLocation, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0700); /*** 7 ***/
snprintf (sModEXE, 100, "%s", arMods_S_Executable_MININIM[iModShowing]);
sModEXE[strlen (sModEXE) - 4] = '\0'; /*** Remove ".exe" ***/

/*** TODO: Maybe use zip_file_extra_field_get() instead? ***/
snprintf (sToWrite, MAX_WRITE, "chmod +x %s\n", sModEXE);
write (iFd, sToWrite, strlen (sToWrite));
#endif
			if (iFd == -1)
			{
				printf ("[ WARN ] CreateBatchFile(): %s!\n", strerror (errno));
			} else {
				snprintf (sToWrite, MAX_WRITE, "%s%s%s%s\n",
					HERE, sModEXE, sModCheat, sModLevel);
			}
		}

		write (iFd, sToWrite, strlen (sToWrite));
		close (iFd);
	}
}
/*****************************************************************************/
void RunGame (char *sNumber)
/*****************************************************************************/
{
	char sDir[MAX_LOCATION + 2];
	char sPar[MAX_LOCATION + 2];
	pid_t pid;

	snprintf (sDir, MAX_LOCATION, "%s%s%s%s",
		MODS_DIR, SLASH, sNumber, SLASH);

	switch (arMods_I_PoP_Version[iModShowing])
	{
		case 1:
		case 2:
			if (strcmp (arMods_S_Executable_DOSSNES[iModShowing], "") != 0)
			{
				snprintf (sPar, MAX_LOCATION, "%s.bat", PROG_NAME);
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
WarnIfNotExists ("DOSBox\\DOSBox.exe");
chdir (sDir);
ShellExecute (NULL, "open", "..\\..\\DOSBox\\DOSBox.exe",
	sPar, NULL, SW_SHOWNORMAL);
chdir ("..\\..");
#else
WarnIfNotExists (BIN_DOSBOX);
pid = fork();
if (pid == 0)
{
	chdir (sDir);
	execl (BIN_DOSBOX, "dosbox", sPar, (char *)NULL);
	exit (EXIT_NORMAL);
}
#endif
			}
			if (strcmp (arMods_S_Executable_SDLPoP[iModShowing], "") != 0)
			{
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
/*** Here, sPar is the batch file itself. ***/
snprintf (sPar, MAX_LOCATION, "%s_s.bat", PROG_NAME);
chdir (sDir);
ShellExecute (NULL, "open", sPar, NULL, sDir, SW_SHOWNORMAL);
chdir ("..\\..");
#else
snprintf (sPar, MAX_LOCATION, "%s_s.sh", PROG_NAME);
pid = fork();
if (pid == 0)
{
	chdir (sDir);
	execl ("/bin/sh", "sh", sPar, (char *)NULL);
	exit (EXIT_NORMAL);
}
#endif
			}
			if (strcmp (arMods_S_Executable_MININIM[iModShowing], "") != 0)
			{
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
/*** Here, sPar is the batch file itself. ***/
snprintf (sPar, MAX_LOCATION, "%s_m.bat", PROG_NAME);
chdir (sDir);
ShellExecute (NULL, "open", sPar, NULL, sDir, SW_SHOWNORMAL);
chdir ("..\\..");
#else
snprintf (sPar, MAX_LOCATION, "%s_m.sh", PROG_NAME);
pid = fork();
if (pid == 0)
{
	chdir (sDir);
	execl ("/bin/sh", "sh", sPar, (char *)NULL);
	exit (EXIT_NORMAL);
}
#endif
			}
			break;
		case 4:
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
snprintf (sPar, MAX_LOCATION, "\"%s\"",
	arMods_S_Executable_DOSSNES[iModShowing]);
WarnIfNotExists ("ZSNES\\zsnesw.exe");
chdir (sDir);
ShellExecute (NULL, "open", "..\\..\\ZSNES\\zsnesw.exe",
	sPar, NULL, SW_SHOWNORMAL);
chdir ("..\\..");
#else
snprintf (sPar, MAX_LOCATION, "%s",
	arMods_S_Executable_DOSSNES[iModShowing]);
WarnIfNotExists (BIN_ZSNES);
pid = fork();
if (pid == 0)
{
	chdir (sDir);
	execl (BIN_ZSNES, "zsnes", sPar, (char *)NULL);
	exit (EXIT_NORMAL);
}
#endif
			break;
	}
}
/*****************************************************************************/
void WarnIfNotExists (char *sFile)
/*****************************************************************************/
{
	GtkWidget *gDialog;

	if (access (sFile, R_OK) == -1)
	{
		gDialog = gtk_message_dialog_new (GTK_WINDOW (gMainWindow),
			GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, "Could not find \"%s\"!", sFile);
		gtk_dialog_run (GTK_DIALOG (gDialog));
		gtk_widget_destroy (gDialog);
	}
}
/*****************************************************************************/
